---
title: X Python Packages to make your own useful CLI tools
tags: [Python, Programming Languages, Coding, Dev Ops, PowerShell, Ansible, programming, engineer] # only 5
publishstatus: draft
canonicalurl: http://medium.com/posts/camratchford/x_things_for_other_things
---


![relavent_picture](./link_to_free_photo.png)

### There are some tasks that do not have the perfect tool for the job. Typically, you'll make a script to automate the process in the hopes that the task is repeatable with minimal human error. A hastily writen script will only get you so far and your jank homebrew likely won't instil a lot of confidence for future users. Make your Python scripts into a more permanent and professional tool with these X Python packages that each bring something to the table.

## Click

[Click Documentation](https://click.palletsprojects.com/en/)

Click is a CLI abstraction layer that is flexible, extensible, and easy to use. The docs are excellently written with plenty of examples (covering most use-cases).

Features include:
- Nested commands
  - You can invoke `cliapp run` and `cli config` instead of `cliapp --run` and `cliapp --config`
- Automatic help page generation
  - The `--help` option content is dynamically generated from the options and arguments you already declared. This can be overridden for a more verbose/specialized output.
- Parameter validation call-back functions, allowing you to do stuff like asynchronous calls to external services.

The description pulled from the Click documentation:

> Click is a Python package for creating beautiful command line interfaces in a composable way with as little code as necessary. It’s the “Command Line Interface Creation Kit”. It’s highly configurable but comes with sensible defaults out of the box.
> It aims to make the process of writing command line tools quick and fun while also preventing any frustration caused by the inability to implement an intended CLI API.

<br>

Also in this category are the Python standard library module [`argparse`](https://docs.python.org/3/library/argparse.html), the docstring-driven package [`Docopt`](http://docopt.org/), and the type-anotation based [`Typer`](https://typer.tiangolo.com/) (from the creator of FastAPI). Each have their particular own use-cases and niche, however most CLI apps will benefit from any these choices.

You should choose Click over the alternatives if:
  - Your argument parsing needs have outgrown that of the standard library `argparse` module
  - You prefer decorators over docstrings or type anotations.
  - You like your code to be explicit.

<br>

How it's used

<br>

> `cli.py`
```python
import click

from __main__ import main

@click.command
@click.option(
    "-d",
    "--debug",
    help="When active, the stream logger will be active and log level set to DEBUG",
    is_flag=True,
)
def run(debug: bool):
    main(debug)

```


## PyYAML

[PyYAML Documentation](https://pyyaml.org/wiki/PyYAMLDocumentation)

PyYAML is the most popular and widely implemnted Python YAML parser/emiter. If you want your app to read in configuration files, YAML is a popular choice.

YAML uses both Python-style indentation to indicate nesting, and a more compact format that uses `[...]` for lists and `{...}` for dicts. It is a subset of JSON, and is a 1:1 drop-in for data serialization means.

You may also consider loading your config files using [`json`](https://docs.python.org/3/library/json.html#module-json) or [`configparser`](https://docs.python.org/3/library/configparser.html) (for TOML-based configs) . They are both part of the standard library and will serve your needs, but (in my own stubborn opinion) lack the combination of flexibilty and readabilty that YAML has. 

A quick example, showing a simple `Config` object being loaded by YAML.

<br>

> `config.py`
```python
from typing import Literal
from pathlib import Path
import yaml


class Config:
    """
    Config object, to capture and store externally-sourced configuration options.
    """

    date_format: str
    log_level: Literal["DEBUG", "INFO", "WARN", "ERROR", "CRIT"]
    log_path: str
    log_config: dict

    def __init__(self, **kwargs):
        # Set defaults
        self.date_format = "%m-%d-%Y %H:%M:%S"
        self.log_level = "INFO"
        self.log_path = ""
        # Write parameters to class attrs
        self.__dict__.update({k: v for k, v in kwargs.items() if k != 'self'})

    @classmethod
    def from_dict(cls, loaded_config: dict):
        if loaded_config:
            config = cls(**loaded_config)
            return config

    @classmethod
    def from_yaml(cls, config_file: str):
        """
        :param config_file: A path string where the yaml config file resides
        :return: A Config object populated with the properties contained in the config file
        """
        config_file_path = Path(config_file).resolve()
        if config_file_path.exists() and (
            ".yml" in config_file_path.suffixes or (".yaml" in config_file_path.suffixes)
        ):
            with open(config_file, "r") as file:
                loaded_config = yaml.safe_load(file)

        return cls.from_dict(loaded_config)
```

<br>

And the implementation, combining Click and our PyYaml Config

<br>

> `cli.py`
```python
import click

from config import Config
from __main__ import main

@click.command
@click.option(
    "-c",
    "--config-file",
    type=click.Path(file_okay=True, dir_okay=False, resolve_path=True),
    help="The location of the yaml configuration file",
    required=False,
)
def run(config_file: str):
    config = Config()
    if config_file:
        config = Config.from_file(config_file)
    main(config)

```

## Jinja2

[Jinja2 Documentation](https://jinja.palletsprojects.com/en/)

Jinj2 is a ubiquitous and wildly extensible templating engine for Python applications.

A templating engine has many use-cases:
- Generate files based on system information, environment variables, database values, and web request data:
  - Config files
  - Reports
  - Documentation (markdown)
  - Boilerplate code (pyproject.toml)
  - Dummy data ()
- Injecting variables into server responses:
  - Web server HTML
  - Telnet / SSH
  - Email

 Jinja is used in many popular Python projects such as Ansible, Boilerplate, or Flask. Jinja has also influnced other templating languages such as Handlebars and Django's own templating engine. 

Features include:
- Template inheritance via the 'block' keyword
- Flexible parsing options, operator mapping, and keyword mapping. (Making it possible to use multiple templating engines possible in the same project).
- A plethora of autoescaping options to prevent XSS from untrusted user input.
- Exceptions point to the correct line in the file, making for super easy debugging of templates.
- Extensible marcos, filters, tests, functions. What it does best IMO. 

<br>

Alternatives include Mako, Django, and the Python string object's `.format()` method.

Use Jinja if you are not using Django and need a templating engine that is extensible, safe, and has lots of compatibilty.


Here's a simple implementation that uses a template called `test.md.j2` from a folder `./templates`, compiling the template and outputting it to `./test.md`

```python
from jinja2 import Environment, FileSystemLoader, select_autoescape
output_path = "./test.md"
template_dir = "./templates"

# Some key-value pairs to use as template variables, that could just as easily be from a more exciting data source
global_variables = {
    "title": "Jinja tutorial",
    "description": "A walkthrough for using jinja templates",
    "disclaimer": "This tutorial isn't very complete"
}

env = Environment(
    loader=FileSystemLoader(template_dir),
    autoescape=select_autoescape(),
)

# Adding our key-value pairs to the templating engine environment.
env.globals.update(global_variables)

template = env.get_template("test.md.j2")

with open(output_path, "w") as file:
    file.write(template.render())
```

The template in use

> `test.md.j2`

```jinja
# {{ title }}

{{ description }}

> {{ disclaimer }}
```

It's output

> `test.md`

```markdown
# Jinja tutorial

A walkthrough for using jinja templates

> This tutorial isn't very complete
```